"""Client configuration."""

from .model import (
    WaylayConfig,
    TenantSettings,
    DEFAULT_PROFILE,
    SERVICE_KEY_GATEWAY,
    SERVICE_KEY_ACCOUNTS,
    SERVICE_KEY_API,
)

__all__ = [
    "WaylayConfig",
    "TenantSettings",
    "DEFAULT_PROFILE",
    "SERVICE_KEY_GATEWAY",
    "SERVICE_KEY_ACCOUNTS",
    "SERVICE_KEY_API",
]
